// $Id: cfdisk.v,v 1.2 2003/10/23 05:11:45 hpa Exp $
//
// CompactFlash controller for ABC80
// Designed to be compatible with the standard ABC-DOS controllers
//
// Note: in order for this to not require internal tristate buffers in
// the FPGA, the data bus is split, and the "input" bus will be driven
// to all FF when the card is not selected, so all inputs can be ANDed.

module cfcontroller(
		    reset_n,	// Global reset
		    clk,	// CPU clk
		    
		    cf_a,	// CompactFlash address bus
		    cf_d,	// CompactFlash data bus
		    cf_rdy,	// CompactFlash RDY
		    cf_wait_n,	// CompactFlash WAIT#
		    cf_ce1_n,	// CompactFlash CE1#
		    cf_ce2_n,	// CompactFlash CE2#
		    cf_oe_n,	// CompactFlash OE#
		    cf_we_n,	// CompactFlash WE#
		    cf_reg_n,	// CompactFlash REG#
		    cf_cd1_n,	// CompactFlash card detect

		    abc_do,	// ABC-bus data out (CPU->controller)
		    abc_di,	// ABC-bus data in  (controller->CPU)
		    abc_out_n,	// ABC-bus data out select (OUT 0)
		    abc_cs_n,	// ABC-bus Card Select
		    abc_c1_n,	// ABC-bus Command 1 (OUT 2)
		    abc_c2_n,	// ABC-bus Command 2 (OUT 3)
		    abc_c3_n,	// ABC-bus Command 3 (OUT 4)
		    abc_c4_n,	// ABC-bus Command 4 (OUT 5)
		    abc_inp_n,	// ABC-bus data in select (IN 0)
		    abc_status_n, // ABC-bus status (IN 1)
		    abc_rst_n,	// ABC-bus reset (IN 7)

		    select,	// Selected?
		    debug
		    );
   
   input         reset_n;
   input         clk;
   
   output [10:0] cf_a;
   inout [15:0]  cf_d;
   input         cf_rdy;
   input         cf_wait_n;
   output        cf_ce1_n;
   output        cf_ce2_n;
   output        cf_oe_n;
   output        cf_we_n;
   output        cf_reg_n;
   input         cf_cd1_n;

   input [7:0] 	 abc_do;
   output [7:0]  abc_di;
   input         abc_out_n;
   input         abc_cs_n;
   input 	 abc_c1_n;
   input         abc_c2_n;
   input         abc_c3_n;
   input 	 abc_c4_n;
   input 	 abc_inp_n;
   input 	 abc_status_n;
   input 	 abc_rst_n;

   output 	 select;
   output [8:0]  debug;
   reg [8:0] 	 debug;
   
   // Forward declaration
   wire 	 cf_cpu_wait_n;	// If we should raise WAIT# to CPU

   // Which select code this device uses... select code 0x24 is used
   // for hard disk controllers
   parameter 	 selectcode = 6'h24;
   reg 		 selected;

   assign 	 select = selected; // For external LED, might need hysteresis
   
   // ------------------------------------------------------------------------
   //  Reset (external and internal)
   // ------------------------------------------------------------------------
   wire 	 xreset = (~reset_n | ~abc_rst_n);
   wire 	 ireset = xreset | (selected & ~abc_c3_n);

   // ------------------------------------------------------------------------
   //  Select logic
   // ------------------------------------------------------------------------

   always @(posedge xreset or posedge clk)
     begin
	if ( xreset )
	  selected <= 0;
	else
	  if ( ~abc_cs_n )
	    selected <= (abc_do[5:0] == selectcode);
     end

   // ------------------------------------------------------------------------
   //  Controller CPU
   // ------------------------------------------------------------------------

   wire 	 cpu_m1_n;
   wire 	 cpu_iorq_n;
   wire 	 cpu_mreq_n;
   wire 	 cpu_rd_n;
   wire 	 cpu_wr_n;
   wire [15:0] 	 cpu_a;
   wire [7:0] 	 cpu_di;
   wire [7:0] 	 cpu_do;
   reg 		 cpu_int_n;
   wire 	 cpu_nmi_n;
   
   T80s cf_cpu (
		.clk_n ( clk ),
		.reset_n ( ~ireset ),
		.wait_n ( cf_cpu_wait_n ),
		.int_n ( cpu_int_n ),
		.nmi_n ( cpu_nmi_n ),
		.busrq_n ( 1 ),
		.m1_n ( cpu_m1_n ),
		.mreq_n ( cpu_mreq_n ),
		.iorq_n ( cpu_iorq_n ),
		.rd_n ( cpu_rd_n ),
		.wr_n ( cpu_wr_n ),
		.a ( cpu_a ),
		.di ( cpu_di ),
		.do ( cpu_do )
		);

   // Issue NMI if CD1# goes high -> card missing
   assign 	 cpu_nmi_n = ~cf_cd1_n;

   always @(posedge ireset or posedge clk)
     begin
	if ( ireset )
	  cpu_int_n <= 1;
	else
	  if ( selected & ~abc_c1_n )
	    cpu_int_n <= 0;
	  else if ( ~cpu_m1_n & ~cpu_iorq_n ) // INTAK
	    cpu_int_n <= 1;
     end // always @ (posedge ireset or posedge clk)
   
   // ------------------------------------------------------------------------
   //  Memory -- second port used for "DMA" to the main CPU
   // ------------------------------------------------------------------------

   wire [7:0]   memrd;		// Read data to controller CPU
   reg 		write_flag;	// DMA if we should write data to controller RAM
   reg 		read_flag;	// DMA if we just read data from controller RAM
   reg [7:0] 	cpudata_di;	// DMA data to controller memory
   wire [7:0] 	cpudata_do;	// DMA data from controller memory
   reg [10:0] 	cpudata_addr;	// DMA target address

   cfram cfram_inst (
                     .clock ( ~clk ),
                     .address_a ( cpu_a[10:0] ),
                     .data_a ( cpu_do ),
                     .wren_a ( ~cpu_mreq_n & ~cpu_wr_n ),
                     .q_a ( memrd ),
		     .address_b ( cpudata_addr ),
		     .data_b ( cpudata_di ),
		     .wren_b ( write_flag ),
		     .q_b ( cpudata_do )
                     );

   // ------------------------------------------------------------------------
   //  Main CPU interface
   // ------------------------------------------------------------------------

   reg [8:0]  cpudata_ctr;	// Bytes left to DMA
   reg 	      cpudata_dir;	// 0 = out, 1 = inp
   reg [7:0]  aux_status;	// Auxilliary status
   reg [7:0]  main_status;	// Primary status
   wire [7:0] dma_status;	// Controller CPU query DMA engine status
   wire       dma_stat_sel = ~cpu_iorq_n & cpu_m1_n & cpu_a[3] & cpu_a[0]; // IN 25

   reg [7:0]  abc_di;
   always @(*)
     begin
	if ( selected & ~abc_inp_n )
	  if ( cpudata_ctr != 0 && cpudata_dir == 1 )
	    abc_di = cpudata_do;
	  else
	    abc_di = aux_status;
	else if ( selected & ~abc_status_n )
	  abc_di = main_status;
	else
	  abc_di = 8'hFF;
     end // always @ (*)
   
   always @(posedge ireset or posedge clk)
     begin
	if ( ireset )
	  begin
	     cpudata_addr <= 0;
	     cpudata_ctr  <= 0;
	     cpudata_dir  <= 0;
	     write_flag   <= 0;
	     read_flag    <= 0;
	     main_status  <= 0;
	     aux_status   <= 0;
	  end
	else
	  begin
	     write_flag <= 0;
	     read_flag  <= 0;
	     
	     if ( selected & ~abc_out_n )
	       begin
		  if ( cpudata_ctr != 0 && cpudata_dir == 0 )
		    begin
		       cpudata_di <= abc_do;
		       write_flag <= 1;
		    end
	       end // if ( selected & ~abc_out_n )
	     else if ( write_flag )
	       begin
		  // We just completed an OUT cycle
		  cpudata_addr <= cpudata_addr + 1;
		  cpudata_ctr  <= cpudata_ctr  - 1;
		  // Turn off START COMMAND
		  main_status[7] <= 0;
		  // If this was the last byte, turn on busy
		  if ( cpudata_ctr == 1 )
		    main_status[0] <= 0;
	       end
	     
	     if ( selected & ~abc_inp_n )
	       begin
		  if ( cpudata_ctr != 0 && cpudata_dir == 1 )
		    begin
		       // abc_di is generated above
		       read_flag <= 1;
		    end
	       end // if ( selected & ~abc_inp_n )
	     else if ( read_flag )
	       begin
		  // We just completed an INP cycle
		  cpudata_addr <= cpudata_addr + 1;
		  cpudata_ctr  <= cpudata_ctr  - 1;
		  // Turn off START COMMAND
		  main_status[7] <= 0;
		  // If this was the last byte, turn on busy
		  if ( cpudata_ctr == 1 )
		    main_status[0] <= 0;
	       end

	     // Set counters or status based on commands from controller CPU
	     if ( ~cpu_iorq_n & cpu_m1_n & ~cpu_wr_n & cpu_a[3] )
	       begin
		  casex ( cpu_a[2:0] )
		    3'b000: // OUT 8
		      main_status <= cpu_do;
		    3'b001: // OUT 9
		      aux_status  <= cpu_do;
		    //3'b01x: // OUT 10-11
		    //  debug <= cpu_do;
		    3'b100: // OUT 12
		      cpudata_addr[7:0] <= cpu_do;
		    3'b101: // OUT 13
		      begin
			 cpudata_addr[10:8] <= cpu_do[2:0];
			 cpudata_dir <= cpu_do[7];
		      end
		    3'b11x: // OUT 14-15
		      cpudata_ctr <= { cpu_a[0], cpu_do };
		  endcase // casex( cpu_a[2:0] )
	       end

	  end // else: !if( ireset )
     end // always @ (posedge ireset or posedge clk)

   assign dma_status = { 6'b0, cpudata_dir, (cpudata_ctr != 0) };

   // ------------------------------------------------------------------------
   //  CompactFlash interface
   // ------------------------------------------------------------------------

`define CF_IDLE	2'h0		// CF card is idle
`define CF_CE   2'h1		// CE# is asserted
`define CF_WOE  2'h2		// WE#/OE# is asserted
`define CF_HOLD 2'h3		// CE# hold cycle

   reg [1:0]  cf_state;
   wire       cf_sel = ~cpu_iorq_n & cpu_m1_n & ~cpu_a[3];
   wire       cf_aux_sel = ~cpu_iorq_n & cpu_m1_n & cpu_a[3] & ~cpu_a[0]; // IN 24 = CF auxilliary status
   reg [2:0]  cf_a_q;
   reg [7:0]  cf_d_q;
   reg 	      cf_write_q;	// True if write cycle
   reg 	      cf_bsy_latch;	// Has RDY/BSY# been 0 since the last write?
   wire [7:0] cf_do;		// Data out to CPU

   // Only 3 address bits provided
   assign     cf_a[10:3] = 8'h00;
   assign     cf_a[2:0]  = cf_a_q;
   
   assign     cf_ce2_n = 1;	// 8-bit access
   assign     cf_reg_n = 1;	// Common memory access mode

   assign     cf_cpu_wait_n = ~(cf_sel & (cf_state != `CF_HOLD));

   assign     cf_d[7:0]  = (cf_write_q & (cf_state != `CF_IDLE)) ? cf_d_q : 8'bz;
   assign     cf_d[15:8] = 8'bz;
   assign     cf_do = (cf_sel & ~cpu_rd_n) ? cf_d_q : 8'hFF;

   assign     cf_ce1_n = ~(cf_state != `CF_IDLE);
   assign     cf_we_n = ~(cf_write_q & (cf_state == `CF_WOE));
   assign     cf_oe_n = ~(~cf_write_q & (cf_state == `CF_WOE));
   
   always @(posedge ireset or posedge clk)
     begin
	if ( ireset )
	  begin
	     cf_state   <= `CF_IDLE;
	     cf_bsy_latch <= 0;
	  end
	else
	  begin
	     if ( ~cf_rdy )
	       cf_bsy_latch <= 1;
	     
	     case ( cf_state )
	       `CF_IDLE:
		 if ( cf_sel & ~(cpu_rd_n & cpu_wr_n) )
		   begin
		      cf_state <= `CF_CE;
		      cf_a_q <= cpu_a[2:0];
		      cf_d_q <= cpu_do;
		      cf_write_q <= ~cpu_wr_n;
		   end
	       `CF_CE:
		 cf_state <= `CF_WOE;
	       `CF_WOE:
		 if ( cf_wait_n ) // If WAIT# low we need to sit and wait
		   begin
		      cf_state <= `CF_HOLD;
		      if ( cf_write_q )
			cf_bsy_latch <= 0;
		      else
			cf_d_q <= cf_d[7:0];
		   end
	       `CF_HOLD:
		 // Stay in this state until CPU cycle complete
		 if ( ~cf_sel | (cpu_rd_n & cpu_wr_n) )
		   cf_state <= `CF_IDLE;
	     endcase // case( cf_state )
	  end
     end

   wire [7:0] cf_aux_status = { cf_state, 3'b0, ~cf_cd1_n, cf_bsy_latch, cf_rdy };

   // ------------------------------------------------------------------------
   //  Controller CPU data select
   // ------------------------------------------------------------------------

   assign cpu_di =
	  cpu_rd_n ? 8'hFF :
	  ~cpu_mreq_n ? memrd :
	  cf_sel ? cf_do :
	  cf_aux_sel ? cf_aux_status :
	  dma_stat_sel ? dma_status :
	  8'hFF;		// Default read input

   // Debugging
   always @(posedge clk)
     begin
	if ( ~cpu_m1_n & ~cpu_mreq_n )
	  debug <= cpu_a[8:0];
     end
   
endmodule // cfcontroller
